from typing import Any, Dict
from pydantic.main import BaseModel
from pydantic import BaseSettings, env_settings
import yaml
import os


class ConfigError(Exception):
    pass


def yaml_config_settings_source(settings: BaseSettings) -> Dict[str, Any]:
    """
    A simple settings source that loads variables from a YAML file
    at the project's root.

    """
    file_path = settings.__config__.yaml_file
    group = settings.__config__.yaml_group
    with open(file_path,"r") as file:
        config = yaml.load(file, Loader=yaml.FullLoader)

    for subgroup in group.split("."):
        config = config[subgroup]

    return config




class BaseConfig(BaseSettings):
    """Container for the Configuration options
    parsed either from Env or from the yaml file

    Args:
        BaseModel ([type]): [description]

    Returns:
        [type]: [description]
    """

    class Config:
        extra = "ignore"

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                env_settings,
                yaml_config_settings_source,
                file_secret_settings,
            )



    @classmethod
    def from_file(cls, file_path=None, **overrides):
        assert hasattr(cls.__config__,"yaml_group"), "Please specifiy your parent group in your Config Class to access it from a file "
        cls.__config__.yaml_file = file_path
        return cls(**overrides)
        



from herre.config.base import BaseConfig
from enum import Enum
from typing import List, Optional


class GrantType(str, Enum):
    IMPLICIT = "IMPLICIT"
    PASSWORD = "PASSWORD"
    CLIENT_CREDENTIALS = "CLIENT_CREDENTIALS"
    AUHORIZATION_CODE = "AUTHORIZATION_CODE"

class LoggingConfig(BaseConfig):

    stream: bool = False
    file: bool = False
    file_path: str = "logs.txt"
    level: str = "WARNING"

    class Config:
        yaml_group = "logging"
        env_prefix = "logging_"

    


    



def in_notebook():
    try:
        from IPython import get_ipython
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False  
    except ImportError:
        return False






from herre.wrap_qt import wrap_qt_in_loop
from herre.logging import setLogging
from herre.jupy import in_notebook
from herre.grants.code.app import AuthorizationCodeGrant
from herre.grants.backend.app import BackendGrant
from herre.grants.backend.app import BackendGrant
from herre.config.herre import GrantType, HerreConfig
import asyncio 
from threading import Thread
import os
import logging

logger = logging.getLogger(__name__)

class HerreError(Exception):
    pass

def newloop(loop):
    asyncio.set_event_loop(loop)
    loop.run_forever()


class HerreClient:

    def __init__(self,
        config_path = "bergen.yaml",
        username: str = None,
        password: str = None,
        allow_insecure = False,
        force_async = False,
        force_sync = False,
        register = True,
        qt = None,
        **overrides,
    ) -> None:
        """ Creates A Herre Client

        Args:
            config_path (str, optional): [description]. Defaults to "bergen.yaml".
            username (str, optional): [description]. Defaults to None.
            password (str, optional): [description]. Defaults to None.
            allow_insecure (bool, optional): [description]. Defaults to False.
            in_sync (bool, optional): Should we force an in_sync modus if an event loop is already running. Loop will be send to another thread. Defaults to True.

        Raises:
            HerreError: [description]
        """

        self.loop = None
        self.thread_id = None


        self.config_path = config_path

        setLogging(config_path=config_path)

        self.config = HerreConfig.from_file(config_path, **overrides)
        print("Here")
        try:
            loop = asyncio.get_running_loop()
            if loop.is_running():
                logger.info("We are in an already running async loop!")
                if in_notebook():
                    logger.info("We are running in a notebook! Choosing on saved preference!")
                    if self.config.jupyter_sync:
                        thread_loop = asyncio.new_event_loop()
                        t = Thread(target=newloop, args=(thread_loop,))
                        t.start()
                        logger.info("Running in Seperate Thread so that we can use the sync syntax")
                        self.loop = thread_loop
                    else:
                        self.loop = loop
                else:
                    logger.info("We are now running in an async Context")
                    self.loop = loop
            else:
                if qt is not None:
                    logger.info("Wrapping Qt app so that we can call functions asynchronously")
                    self.loop = wrap_qt_in_loop(qt)

                self.loop = loop

        except RuntimeError as e:
            logger.info("There is no event-loop in this thread. Lets create a new One")
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            self.loop = loop

        if self.config.authorization_grant_type == GrantType.CLIENT_CREDENTIALS:
            self.grant = BackendGrant(self.config)

        elif self.config.authorization_grant_type == GrantType.AUHORIZATION_CODE:
            self.grant = AuthorizationCodeGrant(self.config)

        else:
            raise HerreError("Unknown GrantType")

        if register:
            set_current_herre(self)


    
    async def login(self, username: str = None, password: str = None):
        await self.grant.login(username=username, password=password)

    
    async def refresh(self):
        assert self.grant.can_refresh, "Grant is not able to refresh. Please login again"
        await self.grant.refresh()


    @property
    def logged_in(self):
        return self.grant.logged_in

    @property
    def headers(self, **additionals):
        assert self.grant.access_token is not None, "Access token is not set yet, please login?"
        return {"Authorization": f"Bearer {self.grant.access_token}", **additionals}




CURRENT_HERRE = None

def get_current_herre(**kwargs):
    global CURRENT_HERRE
    if not CURRENT_HERRE:
        CURRENT_HERRE = HerreClient(**kwargs)
    return CURRENT_HERRE

def set_current_herre(herre):
    global CURRENT_HERRE
    CURRENT_HERRE = herre


from herre.console.context import console, get_current_console
from herre.grants.base import BaseGrant
from oauthlib.oauth2.rfc6749.clients.backend_application import \
    BackendApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session

class BackendGrant(BaseGrant):
    refreshable = True

    def fetchToken(self, **kwargs):
        auth_client = BackendApplicationClient(client_id=self.config.client_id, scope=" ".join(self.config.scopes))
        oauth_session = OAuth2Session(client=auth_client, scope=self.scope)

        self.token = oauth_session.fetch_token(token_url=self.token_url, client_id=self.config.client_id,
            client_secret=self.config.client_secret, verify=True)


    def refreshToken(self):



        return None





from typing import Any, List, Optional
from pydantic.class_validators import validator
from pydantic.fields import Field
from pydantic.main import BaseModel


# Mixin CLass is not Possible because of BaseModel Metaclass
class GraphQLObject(BaseModel):
    typename: Optional[str] = Field(alias="__typename")

    def __init__(__pydantic_self__, **data: Any) -> None:
        if "__typename" not in data: data["__typename"] = __pydantic_self__.__class__.__name__ 
        super().__init__(**data)

    @validator('typename')
    def typename_matches_class(cls, v):
        if v is None: return None # We are ommiting typechecks if __typename is not explicitly set
        if cls.__name__ == v: return v
        raise ValueError(f"Didn't find correct class {cls.__name__} __typename {v}")

    def dict(self, *args, by_alias=True, **kwargs):
        return super().dict(*args, **{
            **kwargs,
            "by_alias": by_alias,
        })

    
